/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2010 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.property;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.ow2.frascati.tinfi.api.control.IllegalPromoterException;
import org.ow2.frascati.tinfi.api.control.SCAContentController;
import org.ow2.frascati.tinfi.api.control.SCAPropertyController;

/**
 * Implementing the {@link SCAContentController} control interface shared by
 * scaPrimitive and scaComposite components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class SCAPropertyControllerMixin
implements SCAPropertyController {
    
    private Map<String,Object> values = new HashMap<String,Object>();
    private Map<String,Class<?>> types = new HashMap<String,Class<?>>();
    
    /**
     * Property name indexed map of promoters.
     * 
     * @since 0.4.3
     */
    private Map<String,SCAPropertyController> promoters =
        new HashMap<String,SCAPropertyController>();
    
    /**
     * Property name indexed map of property names defined at the level of their
     * promoter.
     * 
     * @since 1.4
     */
    private Map<String,String> promoterPropNames = new HashMap<String,String>();

    
    // -------------------------------------------------------------------------
    // Implementation of the Controller interface
    // -------------------------------------------------------------------------

	public void initFcController( InitializationContext ic )
	throws InstantiationException {
		_super_initFcController(ic);
	}

	
    // -------------------------------------------------------------------------
    // Implementation of the SCAPropertyController interface
    // -------------------------------------------------------------------------

    public void setType( String name, Class<?> type ) {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            types.put(name,type);
        }
        else {
        	String promoterPropName = promoterPropNames.get(name);
            promoter.setType(promoterPropName,type);
        }
    }
    
    public void setValue( String name, Object value )
    throws IllegalArgumentException {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {

        	// Check whether the value is of the correct type
        	if( types.containsKey(name) ) {
        		Class<?> type = types.get(name);
        		boolean isInstance = type.isInstance(value);
        		if( ! isInstance ) {
        			String msg =
        				"The type of the specified value "+value+
        				" is not assignable to the type "+type.getName()+
        				" previously defined with setType().";
        			throw new IllegalArgumentException(msg);
        		}
        	}
        	
        	// Record the new value
        	values.put(name,value);
            
            // Re-inject the value on current content instances
            _super_setValue(name,value);
        }
        else {
        	String promoterPropName = promoterPropNames.get(name);
            promoter.setValue(promoterPropName,value);
        }
    }
    
    public Class<?> getType( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
        	Class<?> type = types.get(name);
        	return type;
        }
        else {
        	String promoterPropName = promoterPropNames.get(name);
        	Class<?> type = promoter.getType(promoterPropName);
        	return type;
        }
    }
    
    public Object getValue( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
        	Object value = values.get(name);
        	return value;
        }
        else {
        	String promoterPropName = promoterPropNames.get(name);
        	Object value = promoter.getValue(promoterPropName);
        	return value;
        }
    }
    
    public boolean containsPropertyName( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
        	boolean b = values.containsKey(name);
        	return b;
        }
        else {
        	String promoterPropName = promoterPropNames.get(name);
        	boolean b = promoter.containsPropertyName(promoterPropName);
        	return b;
        }
    }

    public String[] getPropertyNames() {        
        Set<String> keys = values.keySet();
        String[] names = keys.toArray( new String[keys.size()] );
        return names;
    }
    
    public void setPromoter( String name, SCAPropertyController promoter )
    throws IllegalPromoterException {
        
        promoters.put(name,promoter);
        promoterPropNames.put(name,name);

        if( promoter != null ) {
            SCAPropertyController peer = promoter.getPromoter(name);
            if( peer == this ) {
                String compname = _this_weaveableOptNC.getFcName();
                throw new IllegalPromoterException(compname);
            }
        }
    }
    
    public SCAPropertyController getPromoter( String name ) {
        return promoters.get(name);
    }

    public void setPromoter(
    		String name, SCAPropertyController promoter,
    		String promoterPropertyName )
        throws IllegalPromoterException {
    	
    	setPromoter(name,promoter);
    	promoterPropNames.put(name,promoterPropertyName);
    }
    
    public String getPromoterPropertyName( String name ) {
    	return promoterPropNames.get(name);
    }
    
    
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

	protected abstract void _super_initFcController( InitializationContext ic )
	throws InstantiationException;

    protected abstract void _super_setValue( String name, Object value )
    throws IllegalArgumentException;

	private NameController _this_weaveableOptNC;
}
